/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 * vim:expandtab:shiftwidth=8:tabstop=8:
 *
 * Copyright (C) 2001, 2002 Cluster File Systems, Inc.
 * Copyright (C) 2001 Tacit Networks, Inc. <phil@off.net>
 *
 *   This file is part of InterMezzo, http://www.inter-mezzo.org.
 *
 *   InterMezzo is free software; you can redistribute it and/or
 *   modify it under the terms of version 2 of the GNU General Public
 *   License as published by the Free Software Foundation.
 *
 *   InterMezzo is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with InterMezzo; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <pwd.h>
#include <sys/types.h>
#include <glib.h>
#include <string.h>
#include <errno.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <mntent.h>
#include <dirent.h>

#include <linux/intermezzo_fs.h>
#include <izolib.h>
#include <intersync.h>


#define MTAB "/etc/mtab"
#define FS_INTERMEZZO "intermezzo"
#define FS_VINTERMEZZO "vintermezzo"
#define DEVNAME "/dev/intermezzo"

static int get_fstype (char * mtpt) 
{
        struct mntent *ment;
        FILE *fp;
        int rc = FSTYPE_UNKNOWN;
  
        fp = setmntent(MTAB, "r");
        if (fp == NULL) {
                perror(MTAB);
                return FSTYPE_UNKNOWN;
        }

        while((ment = getmntent(fp)) != NULL) {
                if (strcmp(ment->mnt_dir, mtpt) == 0) {
                        if (strcmp(ment->mnt_type, FS_INTERMEZZO) == 0) {
                                rc = FSTYPE_INTERMEZZO;
                        } else if (strcmp(ment->mnt_type, FS_VINTERMEZZO) == 0) {
                                rc = FSTYPE_VINTERMEZZO;
                        }
                        break;
                }
        }
        endmntent(fp);
        return rc;
}

int is_init_kernel(struct is_cache *cache) 
{
        int fd;
        int rc = 0;

        /* Initialize our interface with the kernel */
        mknod(DEVNAME, S_IFCHR | 0700, makedev(10, 248)); 
        fd = open(DEVNAME, O_RDWR);
        if (fd < 0) {
                CERROR("Error opening %s: %s\n",
                       DEVNAME, g_strerror(errno));
                rc = -1;
                goto out;
        }
        
        if (izo_ioc_set_channel(cache->cache_mtpt_fd, fd) < 0) {
                CERROR("izo_ioc_set_channel(%d,%d) failed: %s\n",
                        cache->cache_mtpt_fd, fd, g_strerror(errno));
                close(fd);
                rc = -1;
                goto out;
        }
        
        if (izo_ioc_setpid(cache->cache_mtpt_fd, cache->cache_channel) < 0) {
                CERROR("izo_ioc_setpid(%d,%d) failed: %s\n",
                       cache->cache_mtpt_fd,
                       cache->cache_channel,
                       g_strerror(errno));
                close(fd);
                rc = -1;
                goto out;
        }
        
        cache->cache_intersync->is_kernel_pfd.fd = fd;
 out:
        return rc;
}

struct intersync *is_intersync_new(char * uuid)
{
        struct intersync *is;

        is = g_new0(struct intersync, 1);
        INIT_LIST_HEAD(&is->is_cache_list);
        INIT_LIST_HEAD(&is->is_cell_list);

        if (uuid_parse(uuid, is->is_local_uuid) != 0)
                is_die("Unable to parse local uuid (\"%s\")\n", uuid);
        
        return is;
}

struct is_cell *cell_new(struct intersync *is, char *name)
{
        struct is_cell *cell;

        cell = g_new0(struct is_cell, 1);
        INIT_LIST_HEAD(&cell->cell_peer_list);
        cell->cell_name = g_strdup(name);
        cell->cell_intersync = is;
        cell->cell_srv_lock = g_mutex_new();
        list_add(&cell->cell_intersync_item, &is->is_cell_list);

        return cell;
}

struct is_cell *is_get_cell(struct intersync *is, char *name)
{
        struct is_cell *cell;

        cell = list_entry(is->is_cell_list.next, struct is_cell, cell_intersync_item);
        return cell; 
}

void cell_destroy(struct is_cell *cell)
{
        list_del(&cell->cell_intersync_item);
        g_free(cell->cell_name);
        g_free(cell->cell_root_fset_name);
        g_mutex_free(cell->cell_srv_lock);
        g_free(cell);
}

static struct is_peer *cell_new_peer(struct is_cell *cell, __u64 ip,  __u64 port,
                                     __u64 my_port,
                                     char peer_type)
{
        struct is_peer *peer;
        g_mutex_lock(cell->cell_srv_lock);
        peer = is_peer_ip_lookup(cell, ip, port);
        if (peer != NULL) {
                CDEBUG(D_INFO, "Found peer with same IP address, different UUID\n");
                is_connection_destroy_all(peer);
                is_release_all_permits(cell->cell_intersync, peer);
                is_peer_destroy_rep(peer);
                g_mutex_unlock(cell->cell_srv_lock);
                return peer;
        }

        peer = is_peer_new(ip, port, my_port, peer_type);
        peer->srv_cell = cell;
        list_add(&peer->srv_servers_item, &cell->cell_peer_list);
        g_mutex_unlock(cell->cell_srv_lock);
        return peer;
}

struct is_peer *cell_new_client(struct is_cell *cell, __u64 ip,  __u64 port,
                                __u64 my_port,
                                uuid_t uuid)
{
        struct is_peer *peer;
        peer = cell_new_peer(cell, ip, port, my_port, PEER_CLIENT);
        memcpy (peer->srv_uuid, uuid, sizeof(peer->srv_uuid));
        return peer;
}


struct is_peer *cell_new_server(struct is_cell *cell, __u64 ip,  __u64 port,
                                __u64 my_port)
{
        struct is_peer *peer;
        peer = cell_new_peer(cell, ip, port, my_port, PEER_SERVER);
        return peer;
}

int is_confirm_connection(struct is_peer *peer) 
{
        struct is_cell *cell = peer->srv_cell;
        if (peer->srv_port > 0) {
                if (is_run_get_uuid(cell->cell_intersync, peer) != 0)
                        return -1;
        } else {
                g_mutex_lock(peer->srv_cxn_mutex);
                peer->srv_connected = 0;
                g_mutex_unlock(peer->srv_cxn_mutex);
        }
        return 0;
}


/* XXX we need to be careful not to make more than one UUID call, 
   and not to get duplicate peers in the list: add a condition variable
*/ 
struct is_peer *
cell_add_server(struct is_cell *cell, char *hostname, __u64 port, __u64 my_port)
{
        struct is_peer *server;
        __u64 ip;
        int rc;

        if (izo_resolve(hostname, &ip) < 0) {
                CERROR("cannot resolve hostname of %s\n", hostname);
                return NULL;
        }

        server = cell_new_server(cell, ip, port, my_port);
        rc = is_confirm_connection(server);
        return server;
}

struct is_replicator *is_replicator_new(struct is_fileset *fset,
                                        struct is_peer *peer)
{
        struct is_replicator *rep = g_new0(struct is_replicator, 1);

        rep->rp_peer = peer;
        rep->rp_fset = fset;
        list_add(&rep->rp_peer_item, &peer->srv_replicator_list);
        list_add(&rep->rp_fset_item, &fset->fset_replicator_list);
        /* XXX is RCVD record needed here? */
        /* Make sure 1 and only 1 reconnect thread runs for each peer on the client. */
        g_mutex_lock(peer->srv_cxn_mutex);
        if (peer->srv_peer_type == 1 && peer->srv_cxn_thread == NULL) {
                /* assume we are connected initially */
                peer->srv_connected = 1;
                peer->srv_cxn_thread = g_thread_create((GThreadFunc)is_peer_reconnect_thread,
                                                       peer, TRUE, NULL);
                if (peer->srv_cxn_thread == NULL) {
                        CERROR("peer: %s reconnect thread could not be started.\n",
                               peer->srv_name);
                }
        }
        g_mutex_unlock(peer->srv_cxn_mutex);
        return rep;
}

struct is_replicator *is_replicator_lookup(struct is_fileset *fset,
                                           struct is_peer *peer)
{
        struct list_head *tmp;

        list_for_each(tmp, &fset->fset_replicator_list) {
                struct is_replicator *rep =
                        list_entry(tmp, struct is_replicator, rp_fset_item);
                if (rep->rp_peer == peer)
                        return rep;
        }
        return NULL;
}

void is_peer_destroy_rep(struct is_peer *peer) 
{
        struct list_head *tmp;
        list_for_each(tmp, &peer->srv_replicator_list) {
                struct is_replicator *rep =
                        list_entry(tmp, struct is_replicator, rp_peer_item);
                CDEBUG(D_INFO, " delete rep for %s\n", rep->rp_fset->fset_name);
                list_del(&rep->rp_peer_item);
                list_del(&rep->rp_fset_item);
                /* causes seg fault - why ?
                g_free(rep);
                */

        }

        return;
}


struct is_fileset *is_fileset_new(struct is_cache *cache, char *name,
                                  char *mtpt, struct is_peer *server,
                                  struct is_fileset *parent)
{
        struct is_fileset *fset;
        struct passwd *pw; 
        int rc;

        pw = getpwnam("intermezzo"); 
        if (!pw)
                g_error("no user intermezzo");

        fset = g_new0(struct is_fileset, 1);

        CDEBUG(D_INFO, "setting up new fileset %s dir %s\n", name, mtpt);
        fset->fset_name = g_strdup(name);
        fset->fset_mtpt = g_strdup(mtpt);
        fset->fset_kml_mutex = g_mutex_new();
        fset->fset_kml_cond = g_cond_new();
        fset->fset_is_fetching_kml = 0;
        fset->fset_permit_holder = NULL;

        if (parent == NULL) {
                fset->fset_parent = NULL;
        } else {
                int flags = FSET_IS_BRANCH;
                rc = izo_ioc_mark_fset(parent->fset_mtpt_fd, ~0, FSET_HAS_BRANCHES );
                if (rc < 0) 
                        g_error("Set HAS_BRANCHES on parent failed:  %s\n",
                                g_strerror(errno));
                if (server)
                        flags |= FSET_FLAT_BRANCH;
                CDEBUG(D_INFO, "setting fset %s %x\n", fset->fset_name, flags);
                rc = izo_ioc_set_fset(cache->cache_mtpt_fd, fset->fset_name, 
                                      flags);
                if (rc < 0 && errno != EEXIST) 
                        g_error("set_fset(%s, %s) failed: (%d) %s\n", fset->fset_mtpt,
                                fset->fset_name, errno, g_strerror(errno));
                /* FIXME: lock the parent */
                list_add(&fset->fset_branch_item, &parent->fset_branch_list);
                parent->fset_nbranch++;
                fset->fset_parent = parent;
        }

        fset->fset_mtpt_fd = open(fset->fset_mtpt, O_RDONLY);
        if (fset->fset_mtpt_fd < 0)
                g_error("open(%s) failed: %s\n", fset->fset_mtpt, g_strerror(errno));

        INIT_LIST_HEAD(&fset->fset_replicator_list);

        fset->fset_nbranch = 0;
        INIT_LIST_HEAD(&fset->fset_branch_list);

        rc = izo_ioc_set_uid(fset->fset_mtpt_fd, pw->pw_uid); 
        if (rc)
                g_error("izo_set_ioc_uid failed %d\n", rc); 

        if (server) {
                fset->fset_server = is_replicator_new(fset, server);
        }

        fset->fset_cache = cache;

        list_add(&fset->fset_cache_item, &cache->cache_fset_list);
#if  0
        /*Is this needed? If so, use a different list_head in fset*/
        if (server)
                list_add(&fset->fset_chain, &server->srv_fset_list);
#endif
        return fset;
}

/* 
 * Look for exisiting filesets and branches, and create data
 * structures for them in the kernel.  A fileset is defined as a
 * directory under .intermezzo that contains an entry "ROOT".  The
 * name of the fileset is the path name after .intermezzo, and not
 * including ROOT.  
 * for example,  /mnt/cache/.intermezzo/rootfset/branch/ROOT
 * becomes rootfset/branch
 *
 * This function is recursive.
 */
struct is_fileset *is_build_fileset(struct is_cache *cache, 
                                    char *dir, 
                                    struct is_fileset *parent)
{
        DIR *dirhandle;
        struct dirent *dirent;
        struct is_fileset *fset = NULL;
        struct stat st;

        dirhandle = opendir(dir);
        if (dirhandle == NULL)
                g_error("opendir(%s) failed: %s\n", dir,
                        g_strerror(errno));

        while ((dirent = readdir(dirhandle))) {
                char * fsetname = NULL, *path = NULL, *mtpt = NULL;

                if (dirent->d_name[0] == '.')
                        continue;
                
                if (strcmp(dirent->d_name, "lost+found") == 0)
                        continue;


                path = g_strdup_printf("%s/%s", dir, dirent->d_name);
                if (stat(path, &st))
                        goto next;
                if (!S_ISDIR(st.st_mode))
                        goto next;

                mtpt = g_strdup_printf("%s/ROOT", path);
                if (stat(mtpt, &st))
                        goto next;
                if (!S_ISDIR(st.st_mode))
                        goto next;
                if (parent) {
                        fsetname = g_strdup_printf("%s/%s", 
                                                   parent->fset_name,
                                                   dirent->d_name);
                } else
                        fsetname = g_strdup(dirent->d_name);
                        
                CDEBUG(D_INFO, "Initializing fileset \"%s\"\n", fsetname);
                fset = is_fileset_new(cache, fsetname, mtpt, NULL, parent);
                if (fset == NULL) {
                        CERROR("Unable to create fileset %s\n", fsetname);
                        return NULL;
                }

                is_build_fileset(cache, path, fset);

        next:
                if(fsetname)
                        g_free(fsetname);
                if(mtpt) 
                        g_free(mtpt);
                if(path)
                        g_free(path);
        }
                
        return fset;
}

/*
 * Going to roll forward past this fileset, so this fileset
 * needs to stop pointing to the root.  
 *
 * FIXME: This function doesn't work, as all the files contain
 * pointers to the old fset.  Backfetches fail because they are done
 * through ROOT, and this function changes ROOT to an empty directory,
 * so this breaks backfetches.  I've commented out the call to this
 * func for now.
 */
#if 0
static  int is_fileset_rollpast(struct is_fileset *fset) 
{
        struct is_cache *cache = fset->fset_cache;
        struct stat st;
        int file_flags, fset_flags, cache_flags;
        int rc;

        if (izo_ioc_get_flags(fset->fset_mtpt_fd, &file_flags, &fset_flags, &cache_flags) <0)
                is_die("get_flags(%s) failed: %s\n", 
                        fset->fset_mtpt, g_strerror(errno));
        if (izo_ioc_clear_fset(fset->fset_mtpt_fd) <0) 
                is_die("clear_fset(%s) failed: %s\n", 
                        fset->fset_mtpt, g_strerror(errno));
        close(fset->fset_mtpt_fd);
        if(lstat(fset->fset_mtpt, &st) == 0 && S_ISLNK(st.st_mode)) {
                if(unlink(fset->fset_mtpt) <0) 
                        is_die("unlink(%s) failed: %s\n", 
                               fset->fset_mtpt, g_strerror(errno));
        }
        if(stat(fset->fset_mtpt, &st)) {
                if(mkdir(fset->fset_mtpt, 755) <0)
                        is_die("mkdir(%s) failed: %s\n", 
                               fset->fset_mtpt, g_strerror(errno));
        }
        if(stat(fset->fset_mtpt, &st) || !S_ISDIR(st.st_mode))
                is_die("%s is not directory", fset->fset_mtpt);
        rc = izo_ioc_set_fset(cache->cache_mtpt_fd, fset->fset_name, fset_flags);
        if (rc < 0 && errno != EEXIST) 
                is_die("set_fset(%s, %s) failed: (%d) %s\n", fset->fset_mtpt,
                        fset->fset_name, errno, g_strerror(errno));
        fset->fset_mtpt_fd = open(fset->fset_mtpt, O_RDONLY);
        if (fset->fset_mtpt_fd < 0)
                is_die("open(%s) failed: %s\n", fset->fset_mtpt, g_strerror(errno));
        return 0;
}
#endif 

/*
 * Create the fset directory, and create ROOT symlink 
 */
static int is_prepare_branch_mtpt(char * izo_dir, char * fsetname, char * mtpt) 
{
        char *dir = g_strdup_printf("%s/%s", izo_dir, fsetname);
        int depth, i;
        char * s;
        struct stat st;

        if (dir == NULL)
                is_die("out of memory");

        if (stat(dir, &st)) {
                if (mkdir(dir, 0755) < 0) 
                        is_die("mkdir(%s) failed: %s\n", 
                               dir, g_strerror(errno));
        } else {
                if (!S_ISDIR(st.st_mode))
                        is_die("%s not a directory\n", dir);
        }

        if (stat(mtpt, &st)) {
                char *symlink_path = NULL;
                /* need to create a relative symlink to the root of the fs, 
                 * do this by counting '/' in fsetname, and add 2 for /.intermezzo/.
                 */
                depth = 2;
                s = fsetname;
                while ((s = strchr(s,'/')) != NULL) {
                        depth++;
                        s++;
                }
                symlink_path = g_malloc(depth * 3 +1); /* depth * '../' */
                symlink_path[0] = '\0';
                for(i = 0; i < depth; i++) {
                        strcat (symlink_path, "../");
                }

                if (symlink(symlink_path, mtpt ) <0)
                        is_die("symlink(%s) failed: %s\n", 
                               mtpt, g_strerror(errno));
                g_free(symlink_path);
        } else {
                if (!S_ISDIR(st.st_mode))
                        is_die("%s not a directory\n", dir);
        }
        g_free(dir);
        return 0;
}

static struct is_fileset *is_client_init_fset(struct is_cache *cache,
                                       struct is_peer *server,
                                       char *izo_dir,
                                       char *fsetname)
{
        struct is_fileset * root_fset = NULL;
        struct intersync *is = cache->cache_intersync;
        struct is_replicator *rep;
        char * name, *s;
               
        /* The fset_name includes the branches, if any.   Example:
         *   rootfset/branch1/subbranch1
         * This code steps the parts of the name and creates
         * an fset structure for each part. It will also
         * create the directories if necessary.
         */
        name = s = g_strdup(fsetname);
        while (s != NULL) {
                char branch_mtpt[4096];
                s = strchr(s+1, '/');
                if (s != NULL) *s = '\0';
                CDEBUG(D_INFO, "Initializing fset: %s\n",
                       name);

                snprintf (branch_mtpt, sizeof(branch_mtpt), "%s/%s/ROOT", 
                          izo_dir, name);

                /* 
                 * The first rootfset is currently created by the kernel. 
                 * For the rest, we need to change the previous fset/ROOT 
                 * to a dir and create the directory with ROOT symlink 
                 * for this branch.
                 * FIXME: changing the previos fset/ROOT to a dir doesn't work.
                 */
                if (root_fset != NULL) {
                        /*is_fileset_rollpast(root_fset);*/
                        is_prepare_branch_mtpt(izo_dir, name, branch_mtpt);
                }

                root_fset = is_fileset_new(cache, name, branch_mtpt, 
                                           server, root_fset);
                /* assume we don't have the permit to start with */
                root_fset->fset_permit_holder = server;
                rep = is_replicator_lookup(root_fset, server);
                        
                if (cache->cache_fstype == FSTYPE_VINTERMEZZO && 
                    izo_ioc_mark_fset(root_fset->fset_mtpt_fd, ~0, FSET_DATA_ON_DEMAND))
                        g_error("Unable to set FSET_DATA_ON_DEMAND: %s (%d).\n",
                                g_strerror(errno), root_fset->fset_mtpt_fd);

                if (is_run_job(is, CONTEXT_JOB_RELEASE_PERMIT, NULL, rep, NULL))
                        g_error("Unable to release kernel permit\n");
                if (is_run_job(is, CONTEXT_JOB_GET_LAST_RCVD, NULL, rep, NULL))
                        g_error("Unable to get server last_rcvd data from the "
                                "kernel\n");
                if (is_run_job(is, CONTEXT_JOB_CHECK_OFFSETS, NULL, rep, NULL))
                        g_error("Initial CHECK_OFFSETS failed.\n");
                if (s != NULL) *s = '/';
        }
        g_free(name);
        return root_fset;
}


struct is_fileset *is_fileset_initialize(struct is_cache *cache,
                                         struct is_peer *server,
                                         char * fsetname)
{
        struct is_fileset *root_fset = NULL;
        char *izo_dir = g_strdup_printf("%s/.intermezzo",
                                        cache->cache_mtpt);
        if (server == NULL) {
                root_fset = is_build_fileset(cache, izo_dir, NULL);
        } else {
                if (!fsetname) {
                        CERROR("no root_fset name, can't initialize fset\n");
                        return NULL;
                }
                server->srv_cell->cell_root_fset_name = fsetname;
                root_fset = is_client_init_fset(cache, server, izo_dir,
                                                server->srv_cell->cell_root_fset_name);
        }
        
        g_free(izo_dir);
        return root_fset;
}

struct is_fileset *is_fileset_lookup(struct intersync *is, char *name)
{
        struct list_head *tmp, *fset_tmp;

        list_for_each(tmp, &is->is_cache_list) {
                struct is_cache *cache = list_entry(tmp, struct is_cache,
                                                    cache_intersync_item);
                list_for_each(fset_tmp, &cache->cache_fset_list) {
                        struct is_fileset *fset =
                                list_entry(fset_tmp, struct is_fileset,
                                           fset_cache_item);
                        if (strcmp(fset->fset_name, name) == 0)
                                return fset;
                }
        }
        return NULL;
}

struct is_fileset *is_fileset_make_branch(struct intersync *is,  char *fsetname)
{
        char *parent_name;
        struct is_fileset *parent_fset, *fset = NULL;
        struct is_cache *cache;
        char * mtpt;

        parent_name = izo_dirname(fsetname);
        if (!parent_name)
                return NULL;
        parent_fset = is_fileset_lookup(is, parent_name);
        if (!parent_fset) {
                CERROR("parent fset %s not found!\n", parent_name);
                goto outf;
                
        }
        cache = parent_fset->fset_cache;
        mtpt = g_strdup_printf("%s/.intermezzo/%s/ROOT",
                                   cache->cache_mtpt, fsetname);
        if(izo_mkpath(mtpt, 0755)) {
                CERROR("Unable to make branch mtpt %s\n", mtpt);
                goto outf;
        }

        fset = is_fileset_new(cache, fsetname, mtpt, NULL, parent_fset);
        if (!fset) {
                CERROR("Unable to create fset\n");
        }
 outf:
        g_free(mtpt);
        g_free(parent_name);
        return fset;
}

struct is_cache *is_cache_new(struct intersync *is, char *mtpt)
{
        struct is_cache *cache;
        int fstype;
        fstype = get_fstype(mtpt);
        if (fstype == FSTYPE_UNKNOWN) {
                CERROR("unable to detect intermezzo filesystem: %s\n", mtpt);
                return NULL;
        }

        cache = g_new0(struct is_cache, 1);
        INIT_LIST_HEAD(&cache->cache_intersync_item);
        INIT_LIST_HEAD(&cache->cache_fset_list);

        list_add(&cache->cache_intersync_item, &is->is_cache_list);
        cache->cache_mtpt = mtpt;
        cache->cache_fstype = fstype;
        cache->cache_mtpt_fd = open(mtpt, O_RDONLY);
        if (cache->cache_mtpt_fd < 0)
                g_error("open(%s) failed: %s\n", cache->cache_mtpt, g_strerror(errno));
        cache->cache_intersync = is;
        return cache;
}

struct is_cache *is_cache_initialize(struct intersync *is, char *mtpt, char *docroot)
{
        struct is_cache *cache;
        int rc;
        int channel;
        cache = is_cache_new(is, mtpt);
        if (cache == NULL) 
                return NULL;

        rc = izo_ioc_get_channel(cache->cache_mtpt, &channel);
        if (rc) {
                CERROR("Cannot get kernel channel on cache \"%s\": %s",
                        cache->cache_mtpt, g_strerror(errno));
                is_cache_destroy(cache);
                return NULL;
        }
        cache->cache_channel = channel;

        CDEBUG(D_INFO, "cache %s ----- channel %d\n", cache->cache_mtpt, 
               cache->cache_channel);

        /* Setup the httpd docroot.  If the user didn't specify a docroot
         * directory, use /var/intermezzo-[channel] */
        if (docroot == NULL)
                cache->cache_docroot = g_strdup_printf("/var/intermezzo-%d",
                                                  cache->cache_channel);
        else
                cache->cache_docroot = docroot;

        return cache;
}

void is_cache_destroy(struct is_cache *cache) 
{
        list_del(&cache->cache_intersync_item);
        g_free(cache);
}

struct is_fileset *is_release_all_permits(struct intersync *is, 
                                         struct is_peer *peer)
{
        struct list_head *tmp, *fset_tmp;

        list_for_each(tmp, &is->is_cache_list) {
                struct is_cache *cache = list_entry(tmp, struct is_cache,
                                                    cache_intersync_item);
                list_for_each(fset_tmp, &cache->cache_fset_list) {
                        struct is_fileset *fset =
                                list_entry(fset_tmp, struct is_fileset,
                                           fset_cache_item);
                        if (fset->fset_permit_holder == peer)
                                fset->fset_permit_holder = NULL;
                }
        }
        return NULL;
}
