/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 * vim:expandtab:shiftwidth=8:tabstop=8:
 */

#ifndef INTERSYNC_H
#define INTERSYNC_H

#include "is-list.h"

#include <netinet/in.h>
#include <izolib.h>
#include <glib.h>

/* 100 is the size used in the e2fsprogs sources, though 37 seems
   exactly right (including a trailing NUL). */
#define UUID_UNPARSED_SIZE 37
#include <uuid/uuid.h>
#include <linux/intermezzo_idl.h>

/* 
 * Default config parameters
 */
#define DEFAULT_SERVER_PORT 370
#define DEFAULT_CLIENT_PORT 2432
#define DEFAULT_LOCAL_IP "*"


# define INTERSYNC_DEBUG
# ifdef INTERSYNC_DEBUG
/* global variables */
extern int intersync_debug;
extern int intersync_print_entry;
extern long intersync_kmemory;
extern long intersync_vmemory;

/* debugging masks */
#  define D_INFO        1
#  define D_KML         2
#  define D_UPCALL      4
#  define D_HTTP        8
#  define D_REINT      16
#  define D_CHAT    1024

#  define CDEBUG(mask, format, a...)                                    \
        do {                                                            \
                if (intersync_debug & mask) {                           \
                        printf("(%s:%s L%d): " format, __FILE__,        \
                               __FUNCTION__, __LINE__ , ## a);          \
                }                                                       \
        } while (0)

#  define ENTRY                                                           \
        if (intersync_print_entry)                                        \
                printf("Process %d entered %s\n", getpid(), __FUNCTION__)

#  define EXIT                                                          \
        if (intersync_print_entry)                                      \
                printk("Process %d leaving %s at %d\n", getpid(),       \
                       __FUNCTION__, __LINE__)

#  define intersync_kmem_inc(ptr, size) intersync_kmemory += (size)
#  define intersync_kmem_dec(ptr, size) intersync_kmemory -= (size)
#  define intersync_vmem_inc(ptr, size) intersync_vmemory += (size)
#  define intersync_vmem_dec(ptr, size) intersync_vmemory -= (size)
# else /* !INTERSYNC_DEBUG */
#  define CDEBUG(mask, format, a...) do {} while (0)
#  define ENTRY do {} while (0)
#  define EXIT do {} while (0)
#  define intersync_kmem_inc(ptr, size) do {} while (0)
#  define intersync_kmem_dec(ptr, size) do {} while (0)
#  define intersync_vmem_inc(ptr, size) do {} while (0)
#  define intersync_vmem_dec(ptr, size) do {} while (0)
# endif /* INTERSYNC_DEBUG */

#ifdef CERROR
#undef CERROR
#endif
#define CERROR(format, a...)                                    \
do {                                                            \
        fprintf(stderr, "(%s:%s L%d): " format, __FILE__, __FUNCTION__, \
               __LINE__ , ## a);                                \
} while (0)

#define is_die(format, a...)                                                    \
do {                                                                            \
        fprintf(stderr, "(%s:%s L%d) ABORT: " format, __FILE__, __FUNCTION__,         \
               __LINE__ , ## a);                                                \
        exit(1);                                                                \
                                                                                \
} while (0)

enum {
  FSTYPE_UNKNOWN     = 0,
  FSTYPE_INTERMEZZO  = 1,
  FSTYPE_VINTERMEZZO = 2,
};

struct is_cache {
        struct list_head cache_intersync_item;
        struct list_head cache_fset_list;
        char *cache_mtpt;
        int cache_mtpt_fd;
        int cache_fstype;
        struct intersync *cache_intersync;
        int cache_channel;
        char *cache_docroot;
};

enum {
        PEER_CLIENT = 0,
        PEER_SERVER = 1,
};

struct is_peer {
        char *srv_name;   /* hostname of server */
        int   srv_port;
        struct sockaddr_in srv_addr;
        uuid_t srv_uuid;
        __u64 srv_ip;
        __u64 srv_local_port;
        char srv_peer_type;
        struct is_cell *srv_cell;

        GMutex *srv_kml_mutex;
        GMutex *srv_conn_mutex;
        __u64 srv_kml_size;
        char srv_go_fetch_pending;

        GMutex *srv_cxn_mutex;
        GCond  *srv_cxn_cond;
        char    srv_connected;
        GThread *srv_cxn_thread;

        struct list_head srv_conn_list;
        struct list_head srv_servers_item;
        struct list_head srv_fset_list;     /* fsets on that peer */
        struct list_head srv_replicator_list;
};

struct is_connection {
        struct list_head conn_srv_item;
        struct httpclient *conn_client;
};

struct is_fileset {
        struct is_replicator *fset_server;
        struct is_peer *fset_permit_holder;
        char *fset_name;
        char *fset_mtpt;
        int fset_mtpt_fd;
        struct is_cache *fset_cache;
        struct is_cell *fset_cell;
        struct list_head fset_cache_item;
        struct list_head fset_replicator_list;
        
        struct is_fileset *fset_parent; /* parent branch; NULL if no parent */
        struct list_head fset_branch_item; /* item on parent's branch list */
        struct list_head fset_branch_list; /* branches of this node */
        int fset_nbranch;       /* nbr of branches, if >0, this node is read-only */

        GMutex *fset_kml_mutex;
        GCond  *fset_kml_cond;
        __u64 fset_remote_kml_size;
        char fset_is_fetching_kml;
};

struct is_replicator { 
        struct is_peer *rp_peer;
        struct list_head rp_peer_item;
        struct is_fileset *rp_fset;
        struct list_head rp_fset_item;
        struct izo_rcvd_rec rp_rcvd_rec;
};


/* we have a context for each type of task we perform:
   - getting KML and reintegrating it
   - revoking permit and syncing up, then granting a permit
   - fetching file data 
   each context goes through an execution cycle:
   - new contexts are started at startup, from signals, from upcalls
   - contexts are suspended when I/O would block
   - contexts are resumed from event handlers for I/O on file descriptors
   each context:
   - has a uri/http_req/http_resp/http_trans_conn
*/
struct is_context { 
        struct httpclient *ctxt_req;

        struct intersync *ctxt_intersync;
        struct is_peer *ctxt_peer;
        struct is_replicator *ctxt_replicator;

        struct is_connection *ctxt_isconn;
        int  ctxt_job;
        void *ctxt_job_data;
};

enum {
        CONTEXT_JOB_GET_CONNECTION,
        CONTEXT_JOB_UUID,
        CONTEXT_JOB_KML,
        CONTEXT_JOB_BACKFETCH,
        CONTEXT_JOB_GET_LAST_RCVD,
        CONTEXT_JOB_GET_PROFILE,
        CONTEXT_JOB_CONNECT,
        CONTEXT_JOB_GET_PERMIT,
        CONTEXT_JOB_RELEASE_PERMIT,
        CONTEXT_JOB_GO_FETCH_KML,
        CONTEXT_JOB_SEND_KMLSIZE,
        CONTEXT_JOB_CHECK_OFFSETS,
        CONTEXT_JOB_MAKE_BRANCH,
        CONTEXT_JOB_BRANCH_UNDO,
        CONTEXT_JOB_BRANCH_REDO,
        CONTEXT_JOB_GET_FILEID,
};

struct is_backfetch_data {
        char *path;
        struct lento_vfs_context *info;
};

struct is_fileid_data {
        char *path;
        char *rec;
        int   reclen;
};

struct is_get_uuid_data {
        __u64 port;              /* Requestor port  */
        int client_flag;        /* Requestor is client */
};

struct is_cell {
        struct list_head cell_intersync_item;
        struct list_head cell_peer_list;
        GMutex *cell_srv_lock;

        char *cell_name;
        __u64 cell_host;
        __u64 cell_port;
        char *cell_root_fset_name;
        struct intersync *cell_intersync;
};

struct intersync {
        struct list_head is_cache_list;
        struct list_head is_cell_list;
        GThreadPool      *is_job_pool;
        GThreadPool      *is_upcall_pool;

        uuid_t           is_local_uuid;

        GPollFD          is_kernel_pfd; /* kernel fd shared by everyone */
};

/* connecting.c */
void is_get_http_conn(struct is_context *);
struct is_peer *is_peer_set_ip(struct is_peer *peer, __u64 ip, __u64 port);
struct is_peer *is_peer_new(__u64 ip, __u64 port,  __u64 my_port, int peer_type);
struct is_peer *is_peer_uuid_lookup(struct is_cell *, uuid_t uuid);
struct is_peer *is_peer_ip_lookup(struct is_cell *, __u64 ip,__u64 port);
int is_lowlevel_connect(struct sockaddr *addr, int nonblock);
struct is_connection *is_get_connection(struct is_peer *, int *err);
void is_put_connection(struct is_peer *, struct is_connection *);
void is_connection_destroy(struct is_connection *);
void is_connection_destroy_all(struct is_peer * server);

/* fileset.c */
int is_init_kernel(struct is_cache *cache);
struct intersync *is_intersync_new(char *uuid);
struct is_cell *is_get_cell(struct intersync *, char *name);
struct is_fileset *is_fileset_new(struct is_cache *, char *name, char *mtpt,
                                  struct is_peer *,
                                  struct is_fileset *parent);
struct is_fileset *is_fileset_make_branch(struct intersync *is,  char *fsetname);
struct is_cache *is_cache_new(struct intersync *, char *mtpt);
struct is_cache *is_cache_initialize(struct intersync *, char *mtpt, char *docroot);
void is_cache_destroy(struct is_cache *cache);
struct is_fileset *is_fileset_lookup(struct intersync *, char *name);
struct is_peer *cell_new_client(struct is_cell *cell, __u64 ip,  __u64 port,
                                __u64 my_port,
                                uuid_t uuid);
int is_confirm_connection(struct is_peer *peer);
struct is_peer *cell_add_server(struct is_cell *, char *hostname,
                                __u64 port, __u64 my_port);

struct is_peer *cell_add_peer(struct is_cell *cell, __u64 ip,  __u64 port,
                              uuid_t uuid);
struct is_peer *cell_add_uuid_peer(struct is_peer *server, __u64 my_port, 
                                   int client_flag);
struct is_peer *cell_add_ip_peer(struct is_cell *cell, __u64 ip, __u64 port);
struct is_cell *cell_new(struct intersync *is, char *name);
struct is_replicator *is_replicator_new(struct is_fileset *, struct is_peer *);
struct is_replicator *is_replicator_lookup(struct is_fileset *,
                                           struct is_peer *);
void is_peer_destroy_rep(struct is_peer *peer);
struct is_fileset *is_fileset_initialize(struct is_cache *, struct is_peer *, 
                                         char *fsetname);
struct is_peer *is_peer_set_uuid(struct is_peer *peer, uuid_t uuid);
struct is_fileset *is_release_all_permits(struct intersync *is, 
                                          struct is_peer *peer);

/* request.c */
int is_get_uuid(struct is_context *ctxt, struct is_get_uuid_data *uuid_data);
int is_wait_for_kml(struct is_fileset *fset);
int is_get_kml(struct is_context *ctxt, __u64 kmlsize);
int is_get_file(struct is_context *ctxt, char *path,
                struct lento_vfs_context *info);
int is_get_fileid(struct is_context *ctxt, struct is_fileid_data *fid_data);
int is_get_profile(struct is_context *ctxt);
int is_get_permit(struct is_context *ctxt);
int is_go_fetch_notify(struct intersync *is, struct is_replicator *rep, __u64 *kmlsize);
int is_go_fetch_kml(struct is_context *ctxt, __u64 kmlsize);
int is_check_offsets(struct is_context *ctxt);
int is_branch_undo(struct is_context *ctxt, char *branch);
int is_branch_redo(struct is_context *ctxt, char *branch);

/* session.c */
int is_run_get_uuid(struct intersync *is, struct is_peer *peer);
int is_init_session(struct intersync *is);
inline struct is_peer *is_ctxt_peer(struct is_context *ctxt);
struct is_context *is_context_new(struct intersync *is);
void is_context_destroy(struct is_context *ctxt);
int is_run_context(struct is_context *ctxt);
void is_run_queued_job(struct is_context *ctxt, gpointer data);
void is_context_send_req(struct is_context *ctxt);
void is_queue_job(struct intersync *is, int job, struct is_peer *,
                  struct is_replicator *, gpointer data);
void is_queue_job_delayed(struct intersync *is, int job, struct is_peer *srv,
                          struct is_replicator *, gpointer data, int delay);
int is_run_job(struct intersync *is, int job, struct is_peer *,
               struct is_replicator *, gpointer data);
int is_send_kmlsize(struct is_context *ctxt);
void is_peer_disconnected(struct is_peer *p);
/* upcall.c */
void is_run_upcall(struct is_context *ctxt, gpointer data2);
void is_peer_reconnect_thread(struct is_peer *p);
void izo_sleep(int secs);

#endif
