/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 * vim:expandtab:shiftwidth=8:tabstop=8:
 *
 * Copyright (C) 2001, 2002 Cluster File Systems, Inc.
 * Copyright (C) 2001 Tacit Networks, Inc. <phil@off.net>
 *
 *   This file is part of InterMezzo, http://www.inter-mezzo.org.
 *
 *   InterMezzo is free software; you can redistribute it and/or
 *   modify it under the terms of version 2 of the GNU General Public
 *   License as published by the Free Software Foundation.
 *
 *   InterMezzo is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with InterMezzo; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <glib.h>
#include <string.h>
#include <errno.h>
#include <netdb.h>
#include <sys/poll.h> 
#include <sys/socket.h> 
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <linux/intermezzo_fs.h>
#include <izolib.h>

#include "intersync.h"

static const char *job_names[]  = {
        "JOB_GET_CONNECTION",
        "JOB_UUID",
        "JOB_KML",
        "JOB_BACKFETCH",
        "JOB_GET_LAST_RCVD",
        "JOB_GET_PROFILE",
        "JOB_CONNECT",
        "JOB_GET_PERMIT",
        "JOB_RELEASE_PERMIT",
        "JOB_GO_FETCH_KML",
        "JOB_SEND_KMLSIZE",
        "JOB_CHECK_OFFSETS",
        "JOB_MAKE_BRANCH",
        "JOB_BRANCH_UNDO",
        "JOB_BRANCH_REDO",
        "JOB_GET_FILEID",
};

inline static const char * is_job_name(int job) 
{
        if (job < 0 || job >= sizeof(job_names))
                return "(Unkown  job id)";
        return job_names[job];
}

inline struct is_peer *is_ctxt_peer(struct is_context *ctxt)
{
        struct is_peer *peer;

        if (ctxt->ctxt_replicator) {
                peer = ctxt->ctxt_replicator->rp_peer;
        } else {
                peer = ctxt->ctxt_peer;
        }
        return peer;
}


/* Startup the thread and job infrastructure. */
int is_init_session(struct intersync *is)
{
        int rc = 0;
        static int initialized;
        struct is_context *ctxt;

        if (initialized == 0) {
                initialized = 1;
                g_thread_init(NULL);
                is->is_job_pool =
                        g_thread_pool_new((GFunc)is_run_queued_job, NULL, 5, 0, NULL);
                is->is_upcall_pool =
                        g_thread_pool_new((GFunc)is_run_upcall, NULL, 5, 0, NULL);

                /* Start up 5 upcall handling threads */
                ctxt = is_context_new(is);
                g_thread_pool_push(is->is_upcall_pool, ctxt, NULL);
                g_thread_pool_push(is->is_upcall_pool, ctxt, NULL);
                g_thread_pool_push(is->is_upcall_pool, ctxt, NULL);
                g_thread_pool_push(is->is_upcall_pool, ctxt, NULL);
                g_thread_pool_push(is->is_upcall_pool, ctxt, NULL);
        } else {
                CERROR("threads already initialized\n");
                rc = -1;
        }
        return rc;
}

int is_run_get_uuid(struct intersync *is, struct is_peer *peer)
{
        struct is_get_uuid_data * uuid_data;
        uuid_data = g_new0(struct is_get_uuid_data, 1);
        uuid_data->port = peer->srv_local_port;
        uuid_data->client_flag = peer->srv_peer_type;
        return is_run_job(is, CONTEXT_JOB_UUID, peer, NULL,  uuid_data);
}

/* FIXME: This function is getting out of hand.  I'll refactor it once the demo
 * pressure is off and I have time to closely examine what is needed. */
#define MAX_RETRY 10
int is_run_context(struct is_context *ctxt)
{
        int rc = 0;
        struct is_peer *peer= is_ctxt_peer(ctxt);
        int retry = MAX_RETRY;

        /* There are two differnt types of jobs, local ioctls and rpcs. 
         * To begin refactoring, I split this into seperate switches
         * with the ioclts first. Each ioctl case returns, so the next
         * switch is only RPCs.
         */
        switch(ctxt->ctxt_job) {
        case CONTEXT_JOB_GET_LAST_RCVD: {
                /* Get last_rcvd data about the server from the kernel.  This
                 * requires that the server's UUID be valid, which is set by a
                 * previous CONTEXT_JOB_UUID */
                struct is_replicator *rep = ctxt->ctxt_replicator;
                rc = izo_ioc_get_rcvd(rep->rp_fset->fset_mtpt_fd,
                                      peer->srv_uuid, &rep->rp_rcvd_rec);
                CDEBUG(D_INFO, "get_last_rcvd: kml offset: %Lu\n",
                       rep->rp_rcvd_rec.lr_remote_offset);
                if (rc < 0) {
                        CDEBUG(D_INFO, "izo_ioc_get_rcvd failed: %s\n",
                                  g_strerror(errno));
                        rc = 0;
                }
                return rc;
        }
        case CONTEXT_JOB_RELEASE_PERMIT: {
                /* We run this code when we want to voluntarily give up the
                 * permit (like when the client starts up, in case it was shut
                 * down while still holding the permit) */
                struct is_replicator *rep = ctxt->ctxt_replicator;
                rc = izo_ioc_revoke_permit(rep->rp_fset->fset_mtpt_fd, NULL);
                if (rc != 0) {
                        CERROR("Unable to revoke permit from kernel.  This "
                                  "should never happen.\n");
                }
                return rc;
        }

        }

        if (ctxt->ctxt_job != CONTEXT_JOB_UUID && !peer->srv_connected) {
                CDEBUG(D_HTTP,"%s: ignored, not connected\n", is_job_name(ctxt->ctxt_job));
                return -1;
        }
        while(retry > 0) {
                retry--;
                ctxt->ctxt_isconn = is_get_connection(peer, NULL);
                if (ctxt->ctxt_isconn == NULL) {
                        CERROR("no connection to server: %s:%d\n", peer->srv_name, 
                               peer->srv_port);
                        /* notify reconnect thread for this peer */
                        is_peer_disconnected(peer);
                        rc = -ENOTCONN;
                        break;
                }
                ctxt->ctxt_req = ctxt->ctxt_isconn->conn_client;
                switch(ctxt->ctxt_job) {
                case CONTEXT_JOB_UUID:
                        rc = is_get_uuid(ctxt, ctxt->ctxt_job_data);
                        break;
                case CONTEXT_JOB_KML:
                        if (ctxt->ctxt_job_data)
                                rc = is_get_kml(ctxt, *((__u64 *)ctxt->ctxt_job_data));
                        else
                                is_die("Missing kmlsize\n");
                        break;
                case CONTEXT_JOB_CHECK_OFFSETS:
                        rc = is_check_offsets(ctxt);
                        break;
                case CONTEXT_JOB_BACKFETCH: {
                        struct is_backfetch_data *bf_data = ctxt->ctxt_job_data;
                        struct lento_vfs_context *info = bf_data->info;

                        CDEBUG(D_INFO, "doing backfetch\n");
                        rc = is_get_file(ctxt, bf_data->path, info);
                        break;
                }
                case CONTEXT_JOB_GET_FILEID: {
                        struct is_fileid_data *fid_data = ctxt->ctxt_job_data;

                        CDEBUG(D_INFO, "get fileid\n");
                        rc = is_get_fileid(ctxt, fid_data);
                        break;
                }
                case CONTEXT_JOB_GET_PROFILE:
                        /* Fetch our profile from the server via a CGI. */
                        rc = is_get_profile(ctxt);
                        break;
                case CONTEXT_JOB_GET_PERMIT:
                        /* Make a permit request CGI call */
                        rc = is_get_permit(ctxt);
                        break;
                case CONTEXT_JOB_GO_FETCH_KML:
                        if (ctxt->ctxt_job_data != NULL)
                                rc = is_go_fetch_kml(ctxt,
                                                     *((__u64 *)ctxt->ctxt_job_data));
                        else
                                rc = is_go_fetch_kml(ctxt, 0);
                        break;
                case CONTEXT_JOB_SEND_KMLSIZE:
                        rc = is_send_kmlsize(ctxt);
                        break;
                case CONTEXT_JOB_BRANCH_UNDO:
                        rc = is_branch_undo(ctxt, ctxt->ctxt_job_data);
                        break;
                case CONTEXT_JOB_BRANCH_REDO:
                        rc = is_branch_redo(ctxt, ctxt->ctxt_job_data);
                        break;
                default:
                        g_error("invalid job id: %d\n", ctxt->ctxt_job);
                        rc = 1;
                }
                /* XXX: should this be + or -?  */
                if (rc == -ENOTCONN || rc == ENOTCONN) {
                        CERROR("%s: disconnection detected (%d)\n", is_job_name(ctxt->ctxt_job), 
                               rc);
                        is_connection_destroy(ctxt->ctxt_isconn);
                        is_peer_disconnected(peer);
                        rc = -ENOTCONN;
                        break;
                } else if (rc) {
                        CERROR("%s failed, rc: %d\n", is_job_name(ctxt->ctxt_job), rc);
                        is_connection_destroy(ctxt->ctxt_isconn);
                        if (rc > 0)
                                retry = 0;
                        /* Give up.  FIXME: This should actually keep trying until it's
                         * exhausted all of the pre-existing connections and tried to 
                         * create one new one */
                } else {
                        is_put_connection(peer, ctxt->ctxt_isconn);
                        retry = 0;
                }
        }
        return rc;

}

/* Is called in a thread when a job is pushed into the job queue. */
void is_run_queued_job(struct is_context *ctxt, gpointer data)
{
        is_run_context(ctxt);
        is_context_destroy(ctxt);
}

struct is_context *is_context_new(struct intersync *is)
{
        struct is_context *ctxt = g_new0(struct is_context, 1);
        if (!ctxt) {
                g_error("out of memory");
                return NULL;
        }

        ctxt->ctxt_intersync = is;

        return ctxt;
}

void is_context_init(struct is_context *ctxt, struct is_replicator *rep,
                     struct is_peer *peer)
{
        ctxt->ctxt_replicator = rep;
        ctxt->ctxt_peer = peer;
}

void is_context_destroy(struct is_context *ctxt)
{
        if(ctxt->ctxt_isconn)
                ctxt->ctxt_isconn = NULL;
        if (ctxt->ctxt_req)
                ctxt->ctxt_req = NULL;
        g_free(ctxt->ctxt_job_data);
        g_free(ctxt);
}

/* Is called for delayed jobs when the timer fires. */
static gboolean _really_queue_job(gpointer _ctxt)
{
        struct is_context *ctxt = _ctxt;
        g_thread_pool_push(ctxt->ctxt_intersync->is_job_pool, ctxt, NULL);
        return FALSE;
}

void is_queue_job(struct intersync *is, int job, struct is_peer *peer,
                  struct is_replicator *rep, gpointer data)
{
        is_queue_job_delayed(is, job, peer, rep, data, 0);
}

void is_queue_job_delayed(struct intersync *is, int job,
                          struct is_peer *peer, struct is_replicator *rep,
                          gpointer data, int delay)
{
        struct is_context *ctxt = is_context_new(is);
        is_context_init(ctxt, rep, peer);

        ctxt->ctxt_job = job;
        ctxt->ctxt_job_data = data;
        if (delay == 0)
                g_thread_pool_push(is->is_job_pool, ctxt, NULL);
        else
                g_timeout_add(delay, _really_queue_job, ctxt);
}

/* Runs a job immediately, in the current thread.  Can be run recursively. */
int is_run_job(struct intersync *is, int job, struct is_peer *peer,
               struct is_replicator *rep, gpointer data)
{
        int rc;
        struct is_context *ctxt = is_context_new(is);
        is_context_init(ctxt, rep, peer); 

        ctxt->ctxt_job = job;
        ctxt->ctxt_job_data = data;
        rc = is_run_context(ctxt);
        is_context_destroy(ctxt);
        return rc;
}

/*
 * Sets peer's status to disconnected and notifies reconnect thread.
 * NOTE: This should notify the kernel for filesets  on this peer 
 */
void is_peer_disconnected(struct is_peer *p)
{
        g_mutex_lock(p->srv_cxn_mutex);
        p->srv_connected = 0;
        if (p->srv_cxn_thread) {
                g_cond_broadcast(p->srv_cxn_cond);
        }
        g_mutex_unlock(p->srv_cxn_mutex);
        return;
}

/*
 * Attept to reconnect to peer.
 */
int is_peer_attempt(struct is_peer *p) 
{
        int rc = -ENOTCONN;
        struct is_connection * conn;;
        
        CDEBUG(D_HTTP, "attempt connection to %s:%d\n", 
               p->srv_name, p->srv_port);
        conn = is_get_connection(p, NULL);
        if (conn != NULL) {
                is_put_connection(p, conn);
                rc = is_run_get_uuid(p->srv_cell->cell_intersync, p);
        }
        return rc;
}

#define MILLIS 1000
void izo_sleep(int secs)
{
        poll(NULL, 0, secs*MILLIS);
}

/* Simply attempt at decaying retry times.  */
#define SMART_WAIT(n)                           \
do {                                            \
        izo_sleep(5 * (((n)%10)+1));   \
} while(0);

void is_peer_reconnect_thread(struct is_peer *p) 
{
        int rc, i;
        struct list_head *item;
        CDEBUG(D_HTTP, "peer: %s:%d reconnect thread started.\n",
               p->srv_name, p->srv_port);

 top:
        g_mutex_lock(p->srv_cxn_mutex);
        while(p->srv_connected)
                g_cond_wait(p->srv_cxn_cond, p->srv_cxn_mutex);
        g_mutex_unlock(p->srv_cxn_mutex);
 retry:
        for (i = 0; is_peer_attempt(p); i++)
                SMART_WAIT(i);
        
        /* for each rep on the peer */
        list_for_each(item, &p->srv_replicator_list) {
                struct intersync *is = p->srv_cell->cell_intersync;
                struct is_replicator *rep = list_entry(item, struct is_replicator,
                                                       rp_peer_item);
                /* this should be refactored into a func */
                if (is_run_job(is, CONTEXT_JOB_RELEASE_PERMIT, NULL, rep, NULL))
                        g_error("Unable to release kernel permit\n");
                if (is_run_job(is, CONTEXT_JOB_GET_LAST_RCVD, NULL, rep, NULL))
                        g_error("Unable to get server last_rcvd data from the "
                                "kernel\n");
                rc = is_run_job(is, CONTEXT_JOB_CHECK_OFFSETS, 
                                NULL, rep, NULL);
                if (rc)
                        goto retry;
        }
        goto top;
}



