/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 * vim:expandtab:shiftwidth=8:tabstop=8:
 *
 *  Copyright (C) Tacit Networks, Inc. <phil@off.net>
 *
 *   This file is part of InterMezzo, http://www.inter-mezzo.org.
 *
 *   InterMezzo is free software; you can redistribute it and/or
 *   modify it under the terms of version 2 of the GNU General Public
 *   License as published by the Free Software Foundation.
 *
 *   InterMezzo is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with InterMezzo; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <glib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <netdb.h>
#include <sys/poll.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <fcntl.h>

#include <izolib.h>

enum profile_states {
        START = 0,
        UUID,
        FSET,
        SERVER,
        SERVER_PORT,
        LOCAL_PORT,
        LOCAL_IP,
};

struct state {
        int state;
        struct is_config * cfg;
};

static void parser_start(GMarkupParseContext *context, const gchar *elem,
                         const gchar **attr_name, const gchar **attr_val,
                         gpointer user_data, GError **error)
{
        struct state * s = (struct state *) user_data;
        
        if (strcmp(elem, "profile") == 0) {
                s->state = START;
        } else if (strcmp(elem, "uuid") == 0) {
                s->state = UUID;
        } else if (strcmp(elem, "fset") == 0) {
                s->state = FSET;
        } else if (strcmp(elem, "server") == 0) {
                s->state = SERVER;
        } else if (strcmp(elem, "server_port") == 0) {
                s->state = SERVER_PORT;
        } else if (strcmp(elem, "local_port") == 0) {
                s->state = LOCAL_PORT;
        } else if (strcmp(elem, "local_ip") == 0) {
                s->state = LOCAL_IP;
        } else {
                /* XXX handle errors */
        }
        return;
}

static void parser_text(GMarkupParseContext *context, const gchar *text,
		 gsize text_len, gpointer user_data, GError **error)
{
        struct state * s = (struct state *) user_data;

        switch(s->state) {
        case UUID: 
                s->cfg->cfg_uuid = g_strdup(text);
                break;
        case FSET: 
                s->cfg->cfg_fsetname = g_strdup(text);
                break;
        case SERVER: 
                s->cfg->cfg_server = g_strdup(text);
                break;
        case SERVER_PORT: 
                s->cfg->cfg_server_port = atol(text);
                break;
        case LOCAL_PORT: 
                s->cfg->cfg_local_port = atol(text);
                break;
        case LOCAL_IP: 
                s->cfg->cfg_local_ip = g_strdup(text);
                break;
        default:
                /*printf("Invalid response: '%s'\n", text);*/
        }
}

static void parser_end(GMarkupParseContext *context, const gchar *elem,
                         gpointer user_data, GError **error)
{
        struct state * s = (struct state *) user_data;

        s->state = START;
}


struct is_config *is_parse_response(char * rsp)
{
        static GMarkupParser parser = {
                start_element: parser_start,
                text: parser_text,
                end_element: parser_end,
        };
        GMarkupParseContext *ctx;
        GError *err;
        struct state * s = g_new0(struct state, 1);
        struct is_config * cfg = g_new0(struct is_config, 1);

        s->cfg = cfg;
        ctx = g_markup_parse_context_new(&parser, 0, s, NULL);
        g_markup_parse_context_parse(ctx, rsp, strlen(rsp), &err);
        g_markup_parse_context_free(ctx);
        g_free(s);
        /* fsetname is required */
        if (cfg->cfg_fsetname == NULL) {
                g_free(cfg);
                cfg = NULL;
        }
        return cfg;
}


/*
 * Read whole file into buffer and parse it 
 */
#define MAX_SIZE 16*1024
struct is_config *read_local_config(char *cache) 
{
        char *file;
        char buffer[MAX_SIZE];
        struct is_config *cfg = NULL;
        FILE *fp;
        int len;

        file = g_strdup_printf("%s/.intermezzo/config", cache);
        if (file == NULL)
                goto out;
        fp = fopen(file, "r");
        if (fp == NULL) 
                goto out;
        /* would be good to pass FILE to is_parse, instead of fixed buffer */
        if((len = fread(buffer, 1, sizeof(buffer)-1, fp)) < 0) {
                printf("%s: %s\n", file, g_strerror(errno));
                goto out;
        }
        buffer[len] = '\0';
        cfg = is_parse_response(buffer);
        if (cfg == NULL) {
                printf("unable to parse config: %s\n", file);
        }
 out:
        if (fp)
                fclose(fp);
        if (file)
                g_free(file);
        return cfg;
}

int write_local_config(struct is_config *cfg)
{
        char *file;
        FILE *fp = NULL;
        int rc = -1;

        file = g_strdup_printf("%s/.intermezzo/config", cfg->cfg_cache);
        if (file == NULL) {
                fprintf (stderr, "out of memory\n");
                goto out;
        }
        fp = fopen(file, "w");
        if (fp == NULL) {
                fprintf (stderr, "%s: %s\n", file,
                         g_strerror(errno));
                goto out;
        }
        fprintf(fp, "<profile>\n");
        fprintf(fp," <uuid>%s</uuid>\n",cfg->cfg_uuid);
        fprintf(fp, " <fset>%s</fset>\n", cfg->cfg_fsetname);
        if (cfg->cfg_server)
                fprintf(fp,  " <server>%s</server>\n", cfg->cfg_server);
        if(cfg->cfg_server_port)
                fprintf(fp, " <server_port>%d</server_port>\n", cfg->cfg_server_port);
        if(cfg->cfg_local_ip)
                fprintf(fp, " <local_ip>%s</local_ip>\n", cfg->cfg_local_ip);
        if(cfg->cfg_local_port)
                fprintf(fp, " <local_port>%d</local_port>\n", cfg->cfg_local_port);
        fprintf(fp, "</profile>\n");
        rc = 0;
        
        /* force config to be readable by all */
        chmod (file, 0644);
 out:
        if (fp)
                fclose(fp);
        if (file)
                g_free(file);
        return rc;
}


static char const default_apache_config []= 
"\n\
Listen %s:%d\n\
Port %d\n\
\n\
User intermezzo\n\
Group intermezzo\n\
\n\
LoadModule alias_module         " MODULE_PREFIX "/mod_alias.so\n\
LoadModule cgi_module           " MODULE_PREFIX "/mod_cgi.so\n\
LoadModule config_log_module    " MODULE_PREFIX "/mod_log_config.so\n\
\n\
ErrorLog \"%s/apache_error_log\"\n\
LogLevel warn\n\
LogFormat \"%%h %%l %%u %%t \\\"%%r\\\" %%>s %%b \\\"%%{Referer}i\\\" \\\"%%{User-Agent}i\\\"\" combined\n\
CustomLog \"%s/apache_access_log\" combined\n\
\n\
PidFile %s/httpd.pid\n\
\n\
<Directory />\n\
 AllowOverride None\n\
</Directory>\n\
\n\
<Directory %s/>\n\
  Options ExecCGI\n\
</Directory>\n\
\n\
ScriptAlias /cgi-bin/ %s/\n\
\n\
<VirtualHost _default_:*>\n\
  DocumentRoot \"%s\"\n\
</VirtualHost>\n\
";

void config_apache_conf(char *configfile, char * ip, int port, char *docroot,
                        char *cgi_dir)
{
        gchar *config;
        int fd;

        fd = open(configfile, O_TRUNC|O_CREAT|O_RDWR, 0644); 
        if (fd<0) { 
                g_error("cannot write %s: %s\n", configfile, strerror(errno));
        }

        config = g_strdup_printf(default_apache_config, ip, port, port, docroot, docroot,
                                 docroot, cgi_dir, cgi_dir, docroot);
        write(fd, config, strlen(config));
        g_free(config);
        close(fd); 
}

