/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 * vim:expandtab:shiftwidth=8:tabstop=8:
 *
 * Copyright (C) 2001 Tacit Networks, Inc. <phil@off.net>
 *
 *   This file is part of InterMezzo, http://www.inter-mezzo.org.
 *
 *   InterMezzo is free software; you can redistribute it and/or
 *   modify it under the terms of version 2 of the GNU General Public
 *   License as published by the Free Software Foundation.
 *
 *   InterMezzo is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with InterMezzo; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Userspace ioctl helpers
 *
 */

#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <fcntl.h>
#include <linux/intermezzo_fs.h>
#include <linux/intermezzo_idl.h>
#include <izolib.h>
#include <uuid/uuid.h>

int izo_ioc_mark(int mark_what, int fd, int and_flags, int or_flags)
{
        struct izo_ioctl_data data;
        char *buf = NULL;
        int rc;

        memset(&data, 0, sizeof(data));

        data.ioc_mark_what = mark_what;
        data.ioc_and_flag = and_flags;
        data.ioc_or_flag = or_flags;

        rc = izo_ioctl_pack(&data, &buf, 0);
        if (rc)
                return rc;
        rc = ioctl(fd, IZO_IOC_MARK, buf);
        free(buf);
        return rc;
}

int izo_ioc_mark_fset(int fd, int and_flags, int or_flags)
{
        return izo_ioc_mark(MARK_FSET, fd, and_flags, or_flags);
}

int izo_ioc_get_flags(int fd, int *file_flags, int *and_flags, int *or_flags)
{
        struct izo_ioctl_data data;
        char *buf = NULL;
        int rc;

        memset(&data, 0, sizeof(data));

        data.ioc_mark_what = MARK_GETFL;
        data.ioc_and_flag = *and_flags;
        data.ioc_or_flag = *or_flags;

        rc = izo_ioctl_pack(&data, &buf, 0);
        if (rc)
                return rc;
        rc = ioctl(fd, IZO_IOC_MARK, buf);

        memcpy(&data, buf, sizeof(data));
        *file_flags = data.ioc_mark_what;
        *and_flags = data.ioc_and_flag;
        *or_flags = data.ioc_or_flag;

        free(buf);
        return rc;
}

int izo_ioc_clear_all_fsets(int fd)
{
        int rc;
        struct izo_ioctl_data data;
        char *buf = NULL;

        memset(&data, 0, sizeof(data));
        rc = izo_ioctl_pack(&data, &buf, 0);
        if (rc)
                return rc;
        rc = ioctl(fd, IZO_IOC_CLEAR_ALL_FSETS, buf);
        free(buf);
        return rc;
}

int izo_ioc_clear_fset(int fd)
{
        int rc;
        struct izo_ioctl_data data;
        char *buf = NULL;

        memset(&data, 0, sizeof(data));
        rc = izo_ioctl_pack(&data, &buf, 0);
        if (rc)
                return rc;
        rc = ioctl(fd, IZO_IOC_CLEAR_FSET, buf);
        free(buf);
        return rc;
}

int izo_ioc_set_fset(int fd, char *fset_name, int flags)
{
        int rc;
        struct izo_ioctl_data data;
        char *buf = NULL;
        char dir[4096];
        char *path;

        path = getcwd(dir, sizeof(dir));
        if (!path) {
                printf("cannot getwd on . : %s\n", strerror(errno));
                return -1;
        }

        if (fchdir(fd)) {
                printf("cannot chdir to fset root: %s\n", strerror(errno));
                return -1;
        }

        memset(&data, 0, sizeof(data));
        data.ioc_inllen1 = strlen(fset_name) + 1;
        data.ioc_inlbuf1 = fset_name;
        data.ioc_flags = flags;

        rc = izo_ioctl_pack(&data, &buf, 0);
        if (!rc)
                rc = ioctl(fd, IZO_IOC_SET_FSET, buf);

        if (chdir(path))
                printf("WARNING: cannot chdir back - call the FBI!\n");

        free(buf);
        return rc;
}

int
izo_ioc_get_rcvd(int fd, uuid_t uuid, struct izo_rcvd_rec *rec)
{
        int rc;
        struct izo_ioctl_data data;
        char *buf = NULL;

        memset(&data, 0, sizeof(data));
        memcpy(data.ioc_uuid, uuid, sizeof(data.ioc_uuid));

        rc = izo_ioctl_pack(&data, &buf, 0);
        if (rc)
                return rc;
        rc = ioctl(fd, IZO_IOC_GET_RCVD, buf);
        if (rc < 0)
                return -errno;

        memcpy(rec, buf, sizeof(*rec));
        free(buf);
        return rc;
}

int
izo_ioc_adjust_lml(int fd, struct lento_vfs_context *info)
{
        int rc;
        struct izo_ioctl_data data;
        char *buf = NULL;

        memset(&data, 0, sizeof(data));
        data.ioc_inlbuf1 = (char *)info;
        data.ioc_inllen1 = sizeof(*info);

        rc = izo_ioctl_pack(&data, &buf, 0);
        if (rc)
                return rc;
        rc = ioctl(fd, IZO_IOC_ADJUST_LML, buf);
        if (rc < 0)
                return -errno;

        free(buf);
        return rc;
}


int
izo_ioc_get_kml_size(int fd, __u64 *kmlsize)
{
        int rc;
        struct izo_ioctl_data data;
        char *buf = NULL;

        memset(&data, 0, sizeof(data));
        rc = izo_ioctl_pack(&data, &buf, 0);
        if (rc)
                return rc;
        rc = ioctl(fd, IZO_IOC_GET_KML_SIZE, buf);
        if (rc != 0)
                return rc;

        memcpy(kmlsize, buf, sizeof(*kmlsize));
        free(buf);
        return rc;
}

int
izo_ioc_connect(int fd, __u64 ip_address, __u64 port, uuid_t uuid, 
                int client_flag)
{
        int rc;
        struct izo_ioctl_data data;
        char *buf = NULL;

        memset(&data, 0, sizeof(data));
        data.ioc_ino = ip_address;
        data.ioc_generation = port;
        memcpy(data.ioc_uuid, uuid, sizeof(data.ioc_uuid));
        data.ioc_flags = client_flag;
        rc = izo_ioctl_pack(&data, &buf, 0);
        if (rc)
                return errno;
        rc = ioctl(fd, IZO_IOC_CONNECT, buf);
        free(buf);
        return rc;
}

int
izo_ioc_go_fetch_kml(int fd, uuid_t uuid, __u64 kmlsize)
{
        int rc;
        struct izo_ioctl_data data;
        char *buf = NULL;

        memset(&data, 0, sizeof(data));
        data.ioc_kmlsize = kmlsize;
        memcpy(data.ioc_uuid, uuid, sizeof(data.ioc_uuid));
        rc = izo_ioctl_pack(&data, &buf, 0);
        if (rc)
                return rc;
        rc = ioctl(fd, IZO_IOC_GO_FETCH_KML, buf);
        free(buf);
        return rc;
}

int
izo_ioc_revoke_permit(int fd, uuid_t uuid)
{
        int rc;
        struct izo_ioctl_data data;
        char *buf = NULL;

        memset(&data, 0, sizeof(data));
        if (uuid) {
                memcpy(data.ioc_uuid, uuid, sizeof(data.ioc_uuid));
                data.ioc_flags = 1;
        }
        rc = izo_ioctl_pack(&data, &buf, 0);
        if (rc) {
                printf("izo_ioctl_pack failed: %d (%s)\n", rc, g_strerror(rc));
                return -rc;
        }
        rc = ioctl(fd, IZO_IOC_REVOKE_PERMIT, buf);
        free(buf);
        if (rc < 0)
                rc = -errno;
        return rc;
}

int
izo_ioc_set_kmlsize(int fd, __u64 kmlsize, uuid_t uuid)
{
        int rc;
        struct izo_ioctl_data data;
        char *buf = NULL;

        memset(&data, 0, sizeof(data));
        memcpy(data.ioc_uuid, uuid, sizeof(data.ioc_uuid));
        data.ioc_kmlsize = kmlsize;
        rc = izo_ioctl_pack(&data, &buf, 0);
        if (rc)
                return rc;
        rc = ioctl(fd, IZO_IOC_SET_KMLSIZE, buf);
        free(buf);
        return rc;
}

int izo_ioc_setpid(int fd, __u32 minor)
{
        int rc;
        struct izo_ioctl_data data;
        char *buf = NULL;

        memset(&data, 0, sizeof(data));
        data.ioc_dev = minor;
        rc = izo_ioctl_pack(&data, &buf, 0);
        if (rc)
                return rc;
        rc = ioctl(fd, IZO_IOC_SET_PID, buf);
        free(buf);
        return rc;
}

int izo_ioc_set_uid(int fd, int uid)
{
        int rc;
        struct izo_ioctl_data data;
        char *buf = NULL;

        memset(&data, 0, sizeof(data));
        data.ioc_uid = uid;
        rc = izo_ioctl_pack(&data, &buf, 0);
        if (rc)
                return rc;
        rc = ioctl(fd, IZO_IOC_SET_IOCTL_UID, buf);
        free(buf);
        return rc;
}

int izo_ioc_set_channel(int fd, int channel)
{
        int rc;
        struct izo_ioctl_data data;
        char *buf = NULL;

        memset(&data, 0, sizeof(data));
        data.ioc_dev = channel;
        rc = izo_ioctl_pack(&data, &buf, 0);
        rc = ioctl(fd, IZO_IOC_SET_CHANNEL, buf);
        free(buf);
        return rc;
}

int izo_ioc_get_channel(char *cache, int *channel)
{
        int rc;
        struct izo_ioctl_data data;
        char *buf = NULL;
        int fd;

        fd = open(cache, O_RDONLY); 
        if (fd == -1) { 
                g_error("Cannot open cache \"%s\": %s", cache,
                                g_strerror(errno)); 
        }

        memset(&data, 0, sizeof(data));
        rc = izo_ioctl_pack(&data, &buf, 0);
        if (rc) { 
                close(fd); 
                return rc;
        }
        rc = ioctl(fd, IZO_IOC_GET_CHANNEL, buf);
        if (rc) { 
                close(fd);
                return rc;
        }
        *channel = ((struct izo_ioctl_data *)buf)->ioc_dev;
        free(buf);
        close(fd); 
        return rc;
}

/* Run on the client system to kick off branch creation */
int
izo_ioc_client_make_branch(int fd, char *tagname, char *branchname)
{
        int rc;
        struct izo_ioctl_data data;
        char *buf = NULL;

        memset(&data, 0, sizeof(data));

        data.ioc_inllen1 = strlen(tagname) + 1;
        data.ioc_inlbuf1 = tagname;
        data.ioc_inllen2 = strlen(branchname) + 1;
        data.ioc_inlbuf2 = branchname;

        rc = izo_ioctl_pack(&data, &buf, 0);
        if (rc)
                return rc;
        rc = ioctl(fd, IZO_IOC_CLIENT_MAKE_BRANCH, buf);
        free(buf);
        return rc;
}

int izo_ioc_server_make_branch(int fd, const char *fsetname)
{
        int rc;
        struct izo_ioctl_data data;
        char *buf = NULL;

        memset(&data, 0, sizeof(data));

        data.ioc_inllen1 = strlen(fsetname) + 1;
        data.ioc_inlbuf1 = (char *)fsetname;

        rc = izo_ioctl_pack(&data, &buf, 0);
        if (rc)
                return rc;
        rc = ioctl(fd, IZO_IOC_SERVER_MAKE_BRANCH, buf);
        free(buf);
        return rc;
}

int izo_ioc_branch_undo(int fd, char *branchname)
{
        int rc;
        struct izo_ioctl_data data;
        char *buf = NULL;

        memset(&data, 0, sizeof(data));

        data.ioc_inllen1 = strlen(branchname) + 1;
        data.ioc_inlbuf1 = branchname;

        rc = izo_ioctl_pack(&data, &buf, 0);
        if (rc)
                return rc;
        rc = ioctl(fd, IZO_IOC_BRANCH_UNDO, buf);
        free(buf);
        return rc;
}

int izo_ioc_branch_redo(int fd, char *branchname)
{
        int rc;
        struct izo_ioctl_data data;
        char *buf = NULL;

        memset(&data, 0, sizeof(data));

        data.ioc_inllen1 = strlen(branchname) + 1;
        data.ioc_inlbuf1 = branchname;

        rc = izo_ioctl_pack(&data, &buf, 0);
        if (rc)
                return rc;
        rc = ioctl(fd, IZO_IOC_BRANCH_REDO, buf);
        free(buf);
        return rc;
}


int izo_ioc_get_fileid(int fd, char *rec, __u32 size,
                       __u64 *ino, __u64 *generation)
{
	struct izo_ioctl_data data;
	int rc;

	izo_ioctl_init(&data);

	data.ioc_plen1 = size;
	data.ioc_pbuf1 = rec;

        rc = ioctl(fd, IZO_IOC_GET_FILEID, &data);
	if (rc == 0) {
                *ino = data.ioc_ino;
                *generation = data.ioc_generation;
        }
	return rc;
}

int izo_ioc_set_fileid(int fd, char *path, __u64 ino, __u64 generation)
{
	struct izo_ioctl_data data;
	int rc;

	izo_ioctl_init(&data);

	data.ioc_ino = ino;
	data.ioc_generation = generation;
        data.ioc_pbuf1 = path;
        data.ioc_plen1 = strlen(path) + 1;

        rc = ioctl(fd, IZO_IOC_SET_FILEID, &data);
	return rc;
}

int
izo_ioc_repstatus(int fd, __u64 client_kmlsize,
                  struct izo_rcvd_rec *lr_client,
                  struct izo_rcvd_rec *rec)
{
	struct izo_ioctl_data data;
        char *buf = NULL;
	int rc;

        memset(&data, 0, sizeof(data));

	data.ioc_kmlsize = client_kmlsize;
	data.ioc_plen1 = sizeof(*lr_client);
	data.ioc_pbuf1 = (char*)lr_client;

        rc = izo_ioctl_pack(&data, &buf, 0);
        if (rc)
                return rc;
        rc = ioctl(fd, IZO_IOC_REPSTATUS, buf);
        if (rc < 0)
                return -errno;

        memcpy(rec, buf, sizeof(*rec));
        free(buf);

	return rc;
}

