/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 * vim:expandtab:shiftwidth=8:tabstop=8:
 *
 *  Copyright (C) Tacit Networks, Inc. <phil@off.net>
 *
 *   This file is part of InterMezzo, http://www.inter-mezzo.org.
 *
 *   InterMezzo is free software; you can redistribute it and/or
 *   modify it under the terms of version 2 of the GNU General Public
 *   License as published by the Free Software Foundation.
 *
 *   InterMezzo is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with InterMezzo; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Walks an InterMezzo cache while doing one of two things:
 *
 * - creating a database of file names, atimes, and sizes for use in cache
 *   purging
 * - creating a series of KML records for step 1 of KML truncation
 *
 * It supports an optional XML config file that will tell it to skip subtrees.
 * That file should contain lines that look something like:
 *
 * <hoard dir="/home/phil" dir="/home/braam" />
 *
 * It might support other options later.
 */

#define _GNU_SOURCE
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <ftw.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <time.h>

#include "inwalker.h"

#define I_PURGE_DB 1
#define I_KML 2

#define size_round(x)  (((x)+3) & ~0x3)

#undef DEBUG

static int operation = 0, kml_fd = -1, record_number = 1;
static GList *hoard = NULL;
static GHashTable *link_hash = NULL;

/* Never returns */
static void shutdown(int error)
{
        /* We can't trust Perl to deliver a SIGCHLD, so print a message about
         * our death. */
        if (error == 0)
                printf("SUCCESS\n");
        else
                printf("FAILURE\n");
        exit(error);
}

static void write_kml(void *kml, guint32 size, guint32 opcode,
                      const struct stat *sb)
{
        struct journal_prefix pre;
        struct journal_suffix post;
        int ret;

        size += sizeof(pre) + sizeof(post);

        pre.len = GUINT32_TO_LE(size);
        pre.version = PRESTO_KML_MAJOR_VERSION | PRESTO_KML_MINOR_VERSION;

        /* ? */
        pre.pid = 0;
        pre.uid = 0;
        pre.fsuid = 0;
        pre.fsgid = 0;

        pre.opcode = GUINT32_TO_LE(opcode);

        /* FIXME: double check that this is the right thing -phil */
        pre.ngroups = 1;
        pre.groups[0] = GUINT32_TO_LE(sb->st_gid);

        ret = write(kml_fd, &pre, sizeof(pre));
        if (ret < 0) {
                fprintf(stderr, "write failed: %s\n", g_strerror(errno));
                return;
        } else if (ret < sizeof(pre)) {
                fprintf(stderr, "short write\n");
                return;
        }

        size -= sizeof(pre) + sizeof(post);
        ret = write(kml_fd, kml, size);
        if (ret < 0)
                fprintf(stderr, "write failed: %s\n", g_strerror(errno));
        else if (ret < size)
                fprintf(stderr, "short write\n");
        size += sizeof(pre) + sizeof(post);

        post.prevrec = 0;
        post.recno = record_number++;
        post.time = time(NULL);
        post.len = size;

        ret = write(kml_fd, &post, sizeof(post));
        if (ret < 0) {
                fprintf(stderr, "write failed: %s\n", g_strerror(errno));
                return;
        } else if (ret < sizeof(post)) {
                fprintf(stderr, "short write\n");
                return;
        }
}

#define logit(DEST, SRC, SIZE) { memcpy((DEST), (SRC), (SIZE)); (DEST) += (SIZE); }

static void kml_close(const char *name, const struct stat *sb,
                      const struct stat *parent)
{
        struct kml_close kml;
        char *buf, *tmp;
        int size;
        guint32 pathlen;

        kml.open_mode = sb->st_mode;
        kml.open_uid = sb->st_uid;
        kml.open_gid = sb->st_gid;
        kml.path = (char *)name;
        kml.new_objectv.pv_mtime = sb->st_mtime;
        kml.new_objectv.pv_ctime = sb->st_ctime;
        kml.new_objectv.pv_size = sb->st_size;
        kml.ino = sb->st_ino;
        kml.generation = 0; /* FIXME */

        pathlen = strlen(name);

        size = sizeof(kml.open_mode) + sizeof(kml.open_uid) +
                sizeof(kml.open_gid) + sizeof(kml.new_objectv) +
                sizeof(kml.ino) + sizeof(kml.generation) +
                sizeof(pathlen) + size_round(pathlen);

        tmp = buf = g_malloc(size);

        logit(tmp, &kml.open_mode,   sizeof(kml.open_mode));
        logit(tmp, &kml.open_uid,    sizeof(kml.open_uid));
        logit(tmp, &kml.open_gid,    sizeof(kml.open_gid));
        logit(tmp, &kml.new_objectv, sizeof(kml.new_objectv));
        logit(tmp, &kml.ino,         sizeof(kml.ino));
        logit(tmp, &kml.generation,  sizeof(kml.generation));
        logit(tmp, &pathlen,         sizeof(pathlen));
        logit(tmp, kml.path,         size_round(pathlen));

        write_kml(buf, size, PRESTO_OP_CLOSE, sb);
        g_free(buf);
}

static void kml_mkdir(const char *name, const struct stat *sb,
                      const struct stat *parent)
{
        struct kml_mkdir kml;
        char *buf, *tmp;
        int size;
        guint32 pathlen;

        kml.path = (char *)name;
        kml.new_objectv.pv_mtime = sb->st_mtime;
        kml.new_objectv.pv_ctime = sb->st_ctime;
        kml.new_objectv.pv_size = sb->st_size;
        kml.old_parentv.pv_mtime = parent->st_mtime;
        kml.old_parentv.pv_ctime = parent->st_ctime;
        kml.old_parentv.pv_size = parent->st_size;
        kml.new_parentv.pv_mtime = parent->st_mtime;
        kml.new_parentv.pv_ctime = parent->st_ctime;
        kml.new_parentv.pv_size = parent->st_size;
        kml.mode = sb->st_mode;
        kml.uid = sb->st_uid;
        kml.gid = sb->st_gid;

        pathlen = strlen(name);

        size = 3 * sizeof(kml.new_objectv) + sizeof(kml.mode) + sizeof(kml.uid) +
                sizeof(kml.gid) + sizeof(pathlen) + size_round(pathlen);

        tmp = buf = g_malloc(size);

        logit(tmp, &kml.new_objectv, sizeof(kml.new_objectv));
        logit(tmp, &kml.old_parentv, sizeof(kml.old_parentv));
        logit(tmp, &kml.new_parentv, sizeof(kml.new_parentv));
        logit(tmp, &kml.mode,        sizeof(kml.mode));
        logit(tmp, &kml.uid,         sizeof(kml.uid));
        logit(tmp, &kml.gid,         sizeof(kml.gid));
        logit(tmp, &pathlen,         sizeof(pathlen));
        logit(tmp, kml.path,         size_round(pathlen));

        write_kml(buf, size, PRESTO_OP_MKDIR, sb);
        g_free(buf);
}

static void kml_mknod(const char *name, const struct stat *sb,
                      const struct stat *parent)
{
        struct kml_mknod kml;
        char *buf, *tmp;
        int size;
        guint32 pathlen;

        kml.path = (char *)name;
        kml.new_objectv.pv_mtime = sb->st_mtime;
        kml.new_objectv.pv_ctime = sb->st_ctime;
        kml.new_objectv.pv_size = sb->st_size;
        kml.old_parentv.pv_mtime = parent->st_mtime;
        kml.old_parentv.pv_ctime = parent->st_ctime;
        kml.old_parentv.pv_size = parent->st_size;
        kml.new_parentv.pv_mtime = parent->st_mtime;
        kml.new_parentv.pv_ctime = parent->st_ctime;
        kml.new_parentv.pv_size = parent->st_size;
        kml.mode = sb->st_mode;
        kml.uid = sb->st_uid;
        kml.gid = sb->st_gid;
        kml.major = major(sb->st_rdev);
        kml.minor = minor(sb->st_rdev);

        pathlen = strlen(name);

        size = 3 * sizeof(kml.new_objectv) + sizeof(kml.mode) + sizeof(kml.uid) +
                sizeof(kml.gid) + sizeof(kml.major) + sizeof(kml.minor) +
                sizeof(pathlen) + size_round(pathlen);

        tmp = buf = g_malloc(size);

        logit(tmp, &kml.new_objectv, sizeof(kml.new_objectv));
        logit(tmp, &kml.old_parentv, sizeof(kml.old_parentv));
        logit(tmp, &kml.new_parentv, sizeof(kml.new_parentv));
        logit(tmp, &kml.mode,        sizeof(kml.mode));
        logit(tmp, &kml.uid,         sizeof(kml.uid));
        logit(tmp, &kml.gid,         sizeof(kml.gid));
        logit(tmp, &kml.major,       sizeof(kml.major));
        logit(tmp, &kml.minor,       sizeof(kml.minor));
        logit(tmp, &pathlen,         sizeof(pathlen));
        logit(tmp, kml.path,         size_round(pathlen));

        write_kml(buf, size, PRESTO_OP_MKNOD, sb);
        g_free(buf);
}

static void kml_symlink(const char *name, char *target, const struct stat *sb,
                        const struct stat *parent)
{
        struct kml_symlink kml;
        char *buf, *tmp;
        int size;
        guint32 pathlen, targetlen;

        kml.sourcepath = (char *)name;
        kml.targetpath = target;
        kml.new_objectv.pv_mtime = sb->st_mtime;
        kml.new_objectv.pv_ctime = sb->st_ctime;
        kml.new_objectv.pv_size = sb->st_size;
        kml.old_parentv.pv_mtime = parent->st_mtime;
        kml.old_parentv.pv_ctime = parent->st_ctime;
        kml.old_parentv.pv_size = parent->st_size;
        kml.new_parentv.pv_mtime = parent->st_mtime;
        kml.new_parentv.pv_ctime = parent->st_ctime;
        kml.new_parentv.pv_size = parent->st_size;
        kml.uid = sb->st_uid;
        kml.gid = sb->st_gid;

        pathlen = strlen(name);
        targetlen = strlen(target);

        size = 3 * sizeof(kml.new_objectv) + sizeof(kml.uid) +
                sizeof(kml.gid) +
                sizeof(pathlen) + size_round(pathlen) +
                sizeof(targetlen) + size_round(targetlen);

        tmp = buf = g_malloc(size);

        logit(tmp, &kml.new_objectv, sizeof(kml.new_objectv));
        logit(tmp, &kml.old_parentv, sizeof(kml.old_parentv));
        logit(tmp, &kml.new_parentv, sizeof(kml.new_parentv));
        logit(tmp, &kml.uid,         sizeof(kml.uid));
        logit(tmp, &kml.gid,         sizeof(kml.gid));
        logit(tmp, &pathlen,         sizeof(pathlen));
        logit(tmp, &targetlen,       sizeof(targetlen));
        logit(tmp, name,             size_round(pathlen));
        logit(tmp, target,           size_round(targetlen));

        write_kml(buf, size, PRESTO_OP_SYMLINK, sb);
        g_free(buf);
}

static void kml_link(const char *name, char *target, const struct stat *sb,
                     const struct stat *parent)
{
        struct kml_link kml;
        char *buf, *tmp;
        int size;
        guint32 pathlen, targetlen;

        kml.sourcepath = (char *)name;
        kml.targetpath = target;
        kml.new_objectv.pv_mtime = sb->st_mtime;
        kml.new_objectv.pv_ctime = sb->st_ctime;
        kml.new_objectv.pv_size = sb->st_size;
        kml.old_parentv.pv_mtime = parent->st_mtime;
        kml.old_parentv.pv_ctime = parent->st_ctime;
        kml.old_parentv.pv_size = parent->st_size;
        kml.new_parentv.pv_mtime = parent->st_mtime;
        kml.new_parentv.pv_ctime = parent->st_ctime;
        kml.new_parentv.pv_size = parent->st_size;

        pathlen = strlen(name);
        targetlen = strlen(target);

        size = 3 * sizeof(kml.new_objectv) +
                sizeof(pathlen) + size_round(pathlen) +
                sizeof(targetlen) + size_round(targetlen);

        tmp = buf = g_malloc(size);

        logit(tmp, &kml.new_objectv, sizeof(kml.new_objectv));
        logit(tmp, &kml.old_parentv, sizeof(kml.old_parentv));
        logit(tmp, &kml.new_parentv, sizeof(kml.new_parentv));
        logit(tmp, &pathlen,         sizeof(pathlen));
        logit(tmp, &targetlen,       sizeof(targetlen));
        logit(tmp, name,             size_round(pathlen));
        logit(tmp, target,           size_round(targetlen));

        write_kml(buf, size, PRESTO_OP_LINK, sb);
        g_free(buf);
}

static void usage(char * name)
{
        fprintf(stderr, "usage: %s [-c config] [-d | -k] <directory>\n\n", name);
        fprintf(stderr, "  -c   : specify an optional configuration file\n");
        fprintf(stderr, "  -d   : build a DB for cache purging\n");
        fprintf(stderr, "  -k   : build a DB for KML truncation\n");
        shutdown(1);
}

static int walk_fn(const char *file, const struct stat *sb, int flag,
                   struct FTW *s)
{
        GList *h;

        /* There's no way, unfortunately, to tell ftw() to skip this whole
         * subtree, only abort.  sigh. */
        for (h = hoard; h != NULL; h = h->next) {
                char *ignore = (char *)h->data;
                if (strncmp(ignore, file, strlen(ignore)) == 0) {
                        fprintf(stderr, "ignoring %s\n", file);
                        return 0;
                }
        }

        if (operation == I_PURGE_DB) {
                /* We only care about regular files */
                if (!S_ISREG(sb->st_mode))
                        return 0;

#ifdef DEBUG
                printf("%ld %ld %s\n",
                       (long)sb->st_atime, (long)sb->st_size, file);
#endif
        } else {
                char *link, *tmp;
                struct stat parent;

                tmp = g_strdup(file);
                tmp[s->base] = '\0';
                if (stat(tmp, &parent) < 0) {
                        fprintf(stderr, "Unable to stat() parent '%s': %s\n",
                                tmp, g_strerror(errno));
                        g_free(tmp);
                        return 0;
                }
                g_free(tmp);

                if (sb->st_nlink > 1) {
                        /* We use a hash table to detect hard links */
                        if ((link = g_hash_table_lookup(link_hash,
                                                        (gpointer)sb->st_ino))
                            != NULL) {
                                /* Hard link detected */
                                kml_link(file, link, sb, &parent);
                                return 0;
                        }

                        g_hash_table_insert(link_hash, (gpointer)sb->st_ino,
                                            strdup(file));
                }

                if (S_ISREG(sb->st_mode)) {
                        kml_close(file, sb, &parent);
                } else if (S_ISDIR(sb->st_mode)) {
                        kml_mkdir(file, sb, &parent);
                } else if (S_ISCHR(sb->st_mode) ||
                           S_ISBLK(sb->st_mode)) {
                        kml_mknod(file, sb, &parent);
                } else if (S_ISLNK(sb->st_mode)) {
                        char tmp[4096];
                        int ret;
                        ret = readlink(file, tmp, sizeof(tmp) - 1);
                        if (ret < 0) {
                                fprintf(stderr, "unable to readlink(%s): %s\n",
                                        file, g_strerror(errno));
                        } else {
                                tmp[ret] = '\0';
                        
                                kml_symlink(file, tmp, sb, &parent);
                        }
                }

#ifdef DEBUG
                printf("KML: %ld %ld %s\n",
                       (long)sb->st_atime, (long)sb->st_size, file);
#endif
        }
        return 0;
}

static void parser_start(GMarkupParseContext *context, const gchar *elem,
                         const gchar **attr_name, const gchar **attr_val,
                         gpointer user_data, GError **error)
{
        int i;

        if (strcasecmp(elem, "hoard") != 0) {
                fprintf(stderr, "Ignoring invalid element: %s\n", elem);
                return;
        }

        for (i = 0; attr_name[i] != 0; i++) {
                const char *a = attr_name[i], *v = attr_val[i];

                if (strcasecmp(a, "dir") != 0) {
                        fprintf(stderr, "Ignoring invalid attribute: %s\n", a);
                        return;
                }

                hoard = g_list_prepend(hoard, strdup(v));
        }
}

static int parse_configuration(char *file)
{
        GMarkupParseContext *ctx;
        GMarkupParser parser = {
                start_element: parser_start
        };
        int fd, r;
        char buf[1024];
        GError *err;

        fd = open(file, O_RDONLY);
        if (fd < 0) {
                fprintf(stderr, "Cannot open '%s': %s\n", file,
                        strerror(errno));
                return -1;
        }

        ctx = g_markup_parse_context_new(&parser, 0, NULL, NULL);

        while ((r = read(fd, buf, sizeof(buf))) > 0) {
                g_markup_parse_context_parse(ctx, buf, r, &err);
        }
        g_markup_parse_context_free(ctx);

        return 0;
}

int main(int argc, char *argv[])
{
        int ret;
        char *filename, *configfile = NULL, *kml_file;

        while ((ret = getopt(argc, argv, "c:dk:")) != -1) {
                if (ret == '?' || ret == ':')
                        usage(argv[0]);

                switch(ret) {
                case 'c':
                        configfile = optarg;
                        break;
                case 'd':
                        operation = I_PURGE_DB;
                        break;
                case 'k':
                        kml_file = optarg;
                        operation = I_KML;
                        break;
                }
        }

        if (operation == 0)
               usage(argv[0]);

        if (optind != argc - 1)
                usage(argv[0]);

        filename = argv[optind++];

        if (configfile != NULL && parse_configuration(configfile))
                shutdown(1);

        hoard = g_list_prepend(hoard, g_strdup_printf("%s/.intermezzo",
                                                      filename));

#ifdef DEBUG
        {
                GList *h;
                for (h = hoard; h != NULL; h = h->next)
                        fprintf(stderr, "DEBUG: hoarding %s\n",
                                (char *) h->data);
        }
#endif

        if (operation == I_KML) {
                link_hash = g_hash_table_new(NULL, NULL);
                if (link_hash == NULL)
                        abort();
                kml_fd = open(kml_file, O_CREAT|O_TRUNC|O_RDWR, S_IRWXU|S_IRWXG);
                if (kml_fd < 0) {
                        fprintf(stderr, "unable to open '%s': %s\n", kml_file,
                                g_strerror(errno));
                        shutdown(1);
                }

                /* The KML, of course, wants totally relative pathnames, so to
                 * make this simple just chroot() and then do the ftw() */

                if (chroot(filename) < 0) {
                        fprintf(stderr, "chroot(%s) failure: %s\n", filename,
                                g_strerror(errno));
                        shutdown(1);
                }

                if (nftw("/", walk_fn, 30, FTW_MOUNT | FTW_PHYS) != 0)
                        fprintf(stderr, "nftw failed: %s\n", strerror(errno));
        } else {
                if (nftw(filename, walk_fn, 30, FTW_MOUNT | FTW_PHYS) != 0)
                        fprintf(stderr, "nftw failed: %s\n", strerror(errno));
        }

        shutdown(0);
        return 0;
}
